/*
 *  Copyright (C) 2002-2003  Mattia Dongili<dongili@supereva.it>
 *                           George Staikos <staikos@0wned.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "cpufreqd.h"

short int force_exit; /* force exit from the main loop by signal handlers */
short int force_reinit; /* issue a reinitalization */

struct string_list *env_path;
struct string_list *running_programs;
rule *active_rule;
general *configuration;
short int no_daemon;
void* libsys;   /* handle for the loaded shared library  */
int (*scan_system_info)(sys_info *); /* function pointer */

int main (int argc, char *argv[]) {

  struct sigaction signal_action;

  /* init global vars */
  env_path = string_list_new();
  running_programs = string_list_new();
  active_rule = 0L;
  force_exit = 0;
  force_reinit = 0;
  
  configuration = (general *) malloc(sizeof(general));
  memset(configuration, 0, sizeof(general));
  snprintf(configuration->config_file, 512,  CPUFREQD_CONFIG);
  snprintf(configuration->pidfile, 512, CPUFREQD_PIDFILE);
  
  /* open syslog */
  openlog("cpufreqd", LOG_CONS, LOG_DAEMON);
  
  get_environment_path();
  
  /* parse command line */
  if (read_args(argc,argv)<0)
    return(-1);

  /* check perms */
  if (geteuid() != 0) {
    syslog(LOG_CRIT, "%s: must be run as root.\n", argv[0]);
    return(-1);
  }

  /* check cpufreq interface */
  if (find_cpufreq_interface()<0) {
    syslog(LOG_CRIT, "Unable to find a cpufreq interface, please ensure to have cpufreq enabled in the running kernel.\nExiting.\n");
    return(-1);
  }

  /* read configuration */
  if (parse_config(configuration) < 0) {
    syslog(LOG_CRIT, "Unable to parse configuration file %s.\nExiting.\n", configuration->config_file);
    return(-1);
  }
  /* go background and setup stderr, stdout and stdin */
  if (!no_daemon && daemonize()<0) {
    syslog(LOG_CRIT, "Unable to go to background.\nExiting.\n");
    return(-1);
  }

  syslog(LOG_INFO, "Starting operations.\n");

  /* setup signal handlers */
  sigemptyset(&signal_action.sa_mask);
  sigaddset(&signal_action.sa_mask, SIGTERM);
  sigaddset(&signal_action.sa_mask, SIGINT);
  sigaddset(&signal_action.sa_mask, SIGHUP);
  signal_action.sa_flags = 0;
                             
  signal_action.sa_handler = term_handler;
  sigaction(SIGTERM, &signal_action, 0);

  signal_action.sa_handler = int_handler;
  sigaction(SIGINT, &signal_action, 0);

  signal_action.sa_handler = hup_handler;
  sigaction(SIGHUP, &signal_action, 0);

  /* write pidfile */
  if (write_cpufreqd_pid() < 0) {
    syslog(LOG_CRIT, "Cannot write pidfile %s.\nExiting.\n", configuration->pidfile);
    return -1;
  }
  syslog(LOG_DEBUG, "pid: %d\n\n", getpid());

  /* open libraries */
  if (load_libraries() < 0) {
    syslog(LOG_CRIT, "Cannot open shared libraries.\nExiting.\n");
    return -1;
  }
  
  main_loop();

  close_libraries();

  if (clear_cpufreqd_pid() < 0) {
    syslog(LOG_ERR, "Unable to remove %s.\n", configuration->pidfile);
    return -1;
  }
  
  /* free configuration */
  free_config(configuration);
  
  closelog();
  return(0);
}

/* int load_libraries(void)
 *
 * Open shared libraries
 */
int load_libraries(void) {
  void (*libsys_init)(void); /* init function pointer */
  char libsysname[512];
  const char* error;    /* pointer to an error message, if any */

  switch (configuration->pm_type) {
    case PM_TYPE_ACPI:
       snprintf(libsysname, 512, "libsys_%s.so", "acpi");
       break;
    case PM_TYPE_APM:
    default:
      snprintf(libsysname, 512, "libsys_%s.so", "apm");
  }

  libsys = dlopen(libsysname, RTLD_LAZY);
  if (!libsys) {
    syslog(LOG_ERR, "main_loop(): %s\n", dlerror());
    return -1;
  }

  /* init library */
  libsys_init = dlsym(libsys, "libsys_init");
  error = dlerror();
  if (error) {
    syslog(LOG_ERR, "main_loop(): %s\n", error);
    return -1;
  }
  (*libsys_init)();
  
  /* get function pointer */
  scan_system_info = (int (*)(sys_info*))dlsym(libsys, "scan_system_info");
  error = dlerror();
  if (error) {
    syslog(LOG_ERR, "main_loop(): %s\n", error);
    return -1;
  }
        
  return 0;
              
}

/* void close_libraries(void)
 *
 * Close shared libraries
 */
void close_libraries(void) {
  /* finalize library */
  const char* error;    /* pointer to an error message, if any */
  void (*libsys_fini)() = dlsym(libsys, "libsys_fini");
  error = dlerror();
  if (error) { 
    syslog(LOG_ERR, "main_loop(): %s\n", error);
    return;
  } else
    (*libsys_fini)();
                
  dlclose(libsys);
}

/* void reinit (void)
 *
 * reinit the daemon
 *
 */
void reinit (void) {

  free_config(configuration);
  active_rule = 0L;

  /* close shared library */
  close_libraries();
  
  /* remove pidfile */
  if (clear_cpufreqd_pid() < 0) {
    syslog(LOG_CRIT, "Cannot remove pidfile %s.\nAborting.\n", configuration->pidfile);
    force_exit = 1;
    return;
  }
    
  /* re-read configuration */
  if (parse_config(configuration) < 0) {
    syslog(LOG_CRIT, "Unable to parse configuration file %s.\nAborting.\n", configuration->config_file);
    force_exit = 1;
    return;
  }
  
  /* load shared lib */
  if (load_libraries() < 0) {
    syslog(LOG_CRIT, "Cannot open shared libraries.\nAborting.\n");
    force_exit = 1;
    return;
  }

  /* write pidfile */
  if (write_cpufreqd_pid() < 0) {
    syslog(LOG_CRIT, "Cannot write pidfile %s.\nAborting.\n", configuration->pidfile);
    force_exit = 1;
    return;
  }
  force_reinit = 0;
}

/* void hup_handler(int signo)
 * 
 * SIGUP handler, issues a reinit at the end of the loop.
 *
 */
void hup_handler(int signo) {
  syslog(LOG_INFO, "Caught HUP signal (%s), reloading configuration file.\n", strsignal(signo));

  force_reinit = 1;
}

/* void int_handler(int signo)
 *
 * SIGINT handler.
 */
void int_handler(int signo) {
  syslog(LOG_INFO, "Caught INT signal (%s).\n", strsignal(signo));

  force_exit = 1;
}


/* void term_handler(int signo)
 *
 * SIGTERM handler.
 */
void term_handler(int signo) {
  syslog(LOG_INFO, "Caught TERM signal (%s).\n", strsignal(signo));

  force_exit = 1;
}

/* int main_loop(void)
 *
 * The daemon loop.
 */
int main_loop(void) {
  short int profile_changed = 0; /* to avoid writing a new policy if not really necessary */
  rule *r_temp = 0L;
  sys_info *system = 0L;
  system = (sys_info *)malloc(sizeof(sys_info)); 
  memset(system, 0, sizeof(sys_info));
  
  while (!force_exit) {

    /* scan_system_info */
    if ((*scan_system_info)(system)<0)
      return(-1);

    /* read running programs */
    if (get_running_programs() <= 0) {
      syslog(LOG_WARNING, "main_loop(): WARNING: get_running_programs() returned a negative integer.\n");
    }

    if (system && active_rule) {
      /* check if we need a new rule */
      if (get_rule_hits(system, active_rule) < (MATCH_AC | MATCH_BATTERY | MATCH_PROG)) {

        /* find a valid rule */
        r_temp = find_valid_rule(system, configuration->rules);
        /* check if the new rule is different from the currently active one */
        if (strcmp(r_temp->name, active_rule->name)!=0) {
          /* set profile_changed to avoid writing a new policy if not necessary */
          profile_changed = strcmp(r_temp->prof->name, active_rule->prof->name);
            
          active_rule = r_temp;
          /* set the associated policy */
          /* no need to write it if it is the same profile as before */
          if (profile_changed) {
            if (active_rule && set_policy(active_rule->prof) < 0) {
              return(-1);
            }
          } /* end if profile changed */
        } /* end if rule changed */
      } /* end if rule no more valid */
    } else {
      /* If it the first time we read system status */
      
      /* find a valid rule */
      active_rule = find_valid_rule(system, configuration->rules);

      /* set the associated policy */
      if (active_rule && set_policy(active_rule->prof) < 0)
        return(-1);
    }
    
    if (force_reinit)
      reinit();
    else
      sleep(configuration->poll_interval);
  } /* end loop */

  return(0);
}

/* int get_environment_path(void)
 *
 * Populates the env_path variable
 *
 * Returns 0 on success, -1 otherwise.
 */
int get_environment_path(void) {
        
  char path[512];
  char *entry;
  struct string_node *temp = 0L;
  
  snprintf(path, 512, "%s", getenv("PATH"));
  entry = strtok(path,":");
  
  do {
    temp = string_node_new();
    snprintf(temp->string, 255, "%s", entry);
    string_list_append(env_path, temp);
  } while ((entry = strtok(NULL, ":")));
   
  return 0;
}

/* int write_cpufreqd_pid(void)
 *
 * Writes the pid file.
 *
 * Returns 0 on success, -1 otherwise.
 *
 */
int write_cpufreqd_pid(void) {
  FILE *pid;
  struct stat sb;
  int rc = 0;
  
  /* check if old pidfile is still there */
  rc = stat(configuration->pidfile, &sb);
  if (rc == 0) {
    char oldpid[10];
    pid = fopen(configuration->pidfile, "r");
    /* see if there is a pid already */
    if (fscanf(pid, "%s", oldpid) == 1) {
      FILE *fd;
      char old_pidfile[256];
      char old_cmdline[256];

      snprintf(old_pidfile, 256, "/proc/%s/cmdline", oldpid);
      fd = fopen(old_pidfile, "r");
      /* if the file exists see if there's another cpufreqd process running */
      if (fd && fscanf(fd, "%s", old_cmdline) == 1 && strstr(old_cmdline,"cpufreqd") != NULL) {
        syslog(LOG_ERR, "write_cpufreqd_pid(): the daemon is already running.\n");
        fclose(fd);
        fclose(pid);
        return -1;
      }
    }
    fclose(pid);
  }
  
  pid = fopen(configuration->pidfile, "w");
  if (!pid) {
    syslog(LOG_ERR, "write_cpufreqd_pid(): %s: %s.\n", configuration->pidfile, strerror(errno));
    return -1;
  }

  if (!fprintf(pid, "%d", getpid())) {
    syslog(LOG_ERR, "write_cpufreqd_pid(): cannot write pid %d.\n", getpid());
    fclose(pid);
    clear_cpufreqd_pid();
    return -1;
  }

  fclose(pid);
  return 0;
}

/* int clear_cpufreqd_pid(void)
 *
 * Removes pid file
 *
 * Returns 0 on success, -1 otherwise.
 */
int clear_cpufreqd_pid(void) {
        
  if (unlink(configuration->pidfile) < 0) {
    syslog(LOG_ERR, "clear_cpufreqd_pid(): error while removing %s: %s.\n", configuration->pidfile, strerror(errno));
    return -1;
  }

  return 0;
}

/* int find_cpufreq_interface(void)
 *
 * Checks whether we need to use the 2.4 or 2.5
 * cpufreq iterface and sets the has_sysfs bit
 * accordingly.
 *
 * Returns 0 on success, -1 otherwise
 */
int find_cpufreq_interface(void) {
  struct stat sb;
  int rc;
  
  /* try fisrt with the 2.4 series */
  rc = stat(CPUFREQ_PROC_INTERFACE, &sb);
  if (rc == 0) {
    /*syslog(LOG_DEBUG, "find_cpufreq_interface(): found %s interface.\n", CPUFREQ_PROC_INTERFACE);*/
    configuration->has_sysfs = 0;
    return 0;
  }

  /* try with the 2.5 series */
  rc = stat(CPUFREQ_SYSFS_INTERFACE, &sb);
  if (rc == 0) {
    /*syslog(LOG_DEBUG, "find_cpufreq_interface(): found %s interface.\n", CPUFREQ_SYSFS_INTERFACE);*/
    configuration->has_sysfs = 1;
    return 0;
  }
  
  syslog(LOG_ERR, "find_cpufreq_interface(): no cpufreq interface found.\n");
  return -1;
    
}


/* void set_policy(profile *prof)
 *
 * Sets a new policy through /proc or sysfs interface
 *
 * Returns 0 on success, -1 otherwise
 */
int set_policy(profile *prof) {
  FILE *fp;

  if (!prof)
    return(-1);

  /* /sys/devices/sys/cpu0 */
  if (configuration->has_sysfs) {
    /* governor */
    fp = fopen(CPUFREQ_SYSFS_INTERFACE_POLICY, "w");
    if (!fp) {
      syslog(LOG_ERR, "%s: %s", CPUFREQ_SYSFS_INTERFACE_POLICY, strerror(errno));
      return(-1);
    }
    fprintf(fp, "%s", prof->policy_name);
    fclose(fp);

    /* max speed */
    fp = fopen(CPUFREQ_SYSFS_INTERFACE_MAX, "w");
    if (!fp) {
      syslog(LOG_ERR, "%s: %s", CPUFREQ_SYSFS_INTERFACE_MAX, strerror(errno));
      return(-1);
    }
    fprintf(fp, "%d", prof->max_freq);
    fclose(fp);
                
    /* min speed */
    fp = fopen(CPUFREQ_SYSFS_INTERFACE_MIN, "w");
    if (!fp) {
      syslog(LOG_ERR, "%s: %s", CPUFREQ_SYSFS_INTERFACE_MIN, strerror(errno));
      return(-1);
    }
    fprintf(fp, "%d", prof->min_freq);
    fclose(fp);
    
    syslog(LOG_INFO, "set_policy(): %s %d-%d.\n", prof->policy_name, prof->min_freq, prof->max_freq);
    /*syslog(LOG_WARNING, "Sysfs write yet to be implemented");*/

  } else {

    /* /proc/cpufreq */
    fp = fopen(CPUFREQ_PROC_INTERFACE, "w");
    if (!fp) {
      syslog(LOG_ERR, "%s: %s", CPUFREQ_PROC_INTERFACE, strerror(errno));
      return(-1);
    }

    syslog(LOG_INFO, "set_policy(): 0%%%d%%%d%%%s - profile name: %s.\n",
                     prof->min_freq, prof->max_freq, prof->policy_name, prof->name);
  
    /* TODO: multiple cpu support */
    fprintf(fp, "0%%%d%%%d%%%s", prof->min_freq, prof->max_freq, prof->policy_name);
    fclose(fp);
  }
  return(0);
}


/* rule *find_valid_rule(const sys_info *s, rule *r)
 * 
 * Loops through a rule linked list r and invokes is_rule_valid
 * for each rule.
 *
 * Returns a pointer the the first valid rule or NULL if
 * no valid rules are found.
 */
rule *find_valid_rule(const sys_info *s, rule *r) {

  rule *r_iter = 0L;
  rule *ret = 0L;
  int necessary_hits = (MATCH_AC | MATCH_BATTERY | MATCH_PROG);
  int hits_now = 0L;
  int hits_prev = 0L;
  
  while (necessary_hits > 0) {
    r_iter = r;
    
    while (r_iter != NULL) {
      hits_now = get_rule_hits(s, r_iter);
      /* TODO: add precedence factor to rules having he same score */
      if (hits_now > hits_prev && hits_now >= necessary_hits) {
        hits_prev = hits_now;
        ret = r_iter;
      }
      r_iter = r_iter->next;
    }
    if (ret) {
      /*printf("find_valid_rule(): name %s, hits: %d, necessary: %d\n", 
                       ret->name, hits_prev, necessary_hits);*/
      return ret;
    }
    necessary_hits >>= 1;
  }
  
  syslog(LOG_WARNING, "find_valid_rule(): WARNING: no valid rule found for current"
		  " system state!\n\tbattery: state %s - %d - ac: %s\n",
                  s->has_battery?"present":"absent", 
                  s->battery_percent, 
                  s->ac?"on-line":"off-line");

  return NULL;
}


/* int get_rule_hits(const sys_info *si, const rule *r, int necessary_hits)
 * 
 * Check if the rule r is valid against sytem_info si.
 *
 * Precedence is AC, BATTERY, PROGRAMS
 * 
 * Returns the number of hits the rule matched with, 0 if it doesn't match.
 *
 */
int get_rule_hits(const sys_info *si, const rule *r) {
  int hits = 0, debug = 0;
  
  /* check battery */
  if (si->battery_percent >= r->bat->min && si->battery_percent <= r->bat->max) { 
    hits |= MATCH_BATTERY;
  }
  
  /* check AC */
  if (si->ac == r->ac) {
    hits |= MATCH_AC;
  }
  
  /* check running programs */
  if (r->program_list) {
    char check_string[256];
    struct string_node *pr_iter = 0L, *pr_config = 0L, *path_iter = 0L;
    pr_iter = running_programs->first;
    
    /* go through running programs */
    while (pr_iter && !(MATCH_PROG & hits)) {
      pr_config = r->program_list->first;
      /* go through configured programs */
      while (pr_config && !(MATCH_PROG & hits)) {
        /* first match the configured program name with the running cmdline*/
        if (strcmp(pr_config->string, pr_iter->string) == 0) {
          /*printf( "get_rule_hits(): %s matched requested program \"%s\" against \"%s\"\n",
                  r->name, pr_config->string,  pr_iter->string);*/
          hits |= MATCH_PROG;
        } 
        
        /* try to match configured program name with PATH prefix */
        path_iter = env_path->first;
        while (path_iter && !(MATCH_PROG & hits)) {
          snprintf(check_string, 255, "%s/%s", path_iter->string, pr_config->string);

          if (strcmp(check_string, pr_iter->string) == 0) {
            /*printf("get_rule_hits(): %s matched requested program \"%s\" against \"%s\"\n",
                    r->name, pr_config->string,  pr_iter->string);*/
            hits |= MATCH_PROG;
          }
          path_iter = path_iter->next;
        } /* end env_path loop */
        pr_config = pr_config->next;
      } /* end configured programs loop */
      pr_iter = pr_iter->next;
      debug++;
    } /* end full program list loop */
    /*printf("checked %d processes\n", debug);*/
  }
  /*printf("hits for %s: %d\n", r->name, hits);*/
  return hits;
}

/* int numeric_entry(const struct dirent *d) 
 *
 * Select function for scandir()
 *
 */
int numeric_entry(const struct dirent *d) {
  return isdigit(d->d_name[0]);
}

/* int get_running_programs(void)
 *
 * looks for running programs and fills the 
 * global struct running_programs.
 *
 * Returns the length of the newly created list.
 */
int get_running_programs(void) {
  
  struct dirent **namelist;
  int n=0, ret=0;
  FILE *fp;
  char file[256];
  struct string_node *s_temp = 0L;

  n = scandir("/proc", &namelist, numeric_entry, alphasort);
  if (n < 0)
    syslog(LOG_ERR, "get_running_programs() - scandir: %s\n", strerror(errno));
  else {
    /* reuse the old running_programs string_list */
    s_temp = running_programs->first;
    
    while(n--) {
      snprintf(file, 255, "/proc/%s/cmdline", namelist[n]->d_name);    
      free(namelist[n]);
      fp = fopen(file, "r");
      
      if (!fp) /* probably this process has disappeared while scanning, don't worry */
        continue;

      if (!s_temp) {
        s_temp = string_node_new();
        string_list_append(running_programs, s_temp);
      }
      /* reset string */
      s_temp->string[0] = '\0';
      /* read cmdline */
      if (!fgets(s_temp->string, 255, fp)) {
        fclose(fp);
        /* this entry is not readable, go on */
        continue;
      }
      fclose(fp);
      
      s_temp = s_temp->next;

      ret++;
    }
  }
  free(namelist);
  /*printf("read %d processes\n", ret);*/
  /*printf("freed %d structs\n", string_list_free_sublist(running_programs, s_temp));*/
  return ret;
}

/*
 * Creates a child and detach from tty
 */
int daemonize (void) {

  /*syslog(LOG_DEBUG, "daemonize(): going background, bye.\n");*/

  switch (fork ()) {
    case -1:
      syslog(LOG_ERR, "deamonize(): fork: %s\n", strerror (errno));
      return -1;
    case 0:
      /* child */
      break;
    default:
      /* parent */
      exit (0);
  }

  /* disconnect */
  setsid ();
  umask (0);

  /* set up stdout to log and stderr,stdin to /dev/null */
  if (freopen("/dev/null", "r", stdin) == NULL) {
    syslog(LOG_ERR, "deamonize(): %s: %s.\n", "/dev/null", strerror(errno));
    return -1;
  }
  
  if (freopen("/dev/null", "w", stdout) == NULL) {
    syslog(LOG_ERR, "deamonize(): %s: %s.\n", "/dev/null", strerror(errno));
    return -1;
  }
  
  if (freopen("/dev/null", "w", stderr) == NULL) {
    syslog(LOG_ERR, "deamonize(): %s: %s.\n", "/dev/null", strerror(errno));
    return -1;
  }

  /* get outta the way */
  chdir ("/");
  return 0;
}

/*
 * Reads command line arguments
 *
 */
int read_args (int argc, char *argv[]) {
  
  static struct option long_options[] = {
    { "help", 0, 0, 'h' }, 
    { "version", 0, 0, 'v' }, 
    { "file", 1, 0, 'f' },
    { "no-daemon", 0, 0, 'D' },
    { 0, 0, 0, 0 }, 
  };
  int ch,option_index = 0;

  while ((ch = getopt_long(argc, argv, "hvf:D", long_options, &option_index)) != -1) {
    switch (ch) {
      case '?':
      case 'h':
        print_help(argv[0]);
        return -1;
      case 'v':
        print_version(argv[0]);
        return -1;
      case 'f':
        if (realpath(optarg, configuration->config_file) == NULL) {
          syslog(LOG_ERR, "read_args(): %s: %s.\n", optarg, strerror(errno));
          return -1;
        }
        break;
      case 'D':
        no_daemon = 1;
        break;
      default:
        break;
    }
  }
  return 0;
}

/*
 * Prints program version
 */
void print_version(const char *me) {
  printf("%s version "_CPUFREQD_VERSION_".\n", me);
  printf("Copyright 2002-2003, Mattia Dongili <dongili@supereva.it>\n"
	    	 "                     George Staikos <staikos@0wned.org>\n");
}

/*
 * Prints program help
 *
 */
void print_help(const char *me) {
  printf("Usage: %s [OPTION]...\n\n"
"  -h, --help                   display this help and exit\n"
"  -v, --version                display version information and exit\n"
"  -f, --file                   config file (default: /etc/cpufreqd.conf)\n"
"  -D, --no-daemon              stay in foreground (used to debug)\n"
"\n"
"Report bugs to Mattia Dongili <dongili@supereva.it>.\n", me);
}

